unit sPageControl;
(************************************************************************
sStyle components package
Copyright (c) 2002 Serge Goncharov
                            Ukraine, Odessa
                            http://www.geocities.com/sstylish2002/
                            Serge@HyBce.com
 Version:	1.1, 04-02-2002
*************************************************************************)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, sStyleUtil, sConst, ExtCtrls, sPanel, sGraphUtils, sUtils, ImgList,
  Consts, ComStrs, CommCtrl, dsgnintf;

type
  TsPageControl = class;

  TsTabSheet = class(TsCustomPanel)
  private
    FImageIndex: TImageIndex;
    FPageControl: TsPageControl;
    FTabVisible: Boolean;
    FTabShowing: Boolean;
    FHighlighted: Boolean;
    FOnHide: TNotifyEvent;
    FOnShow: TNotifyEvent;
    function GetPageIndex: Integer;
    function GetTabIndex: Integer;
    procedure SetHighlighted(Value: Boolean);
    procedure SetImageIndex(Value: TImageIndex);
    procedure SetPageControl(APageControl: TsPageControl);
    procedure SetPageIndex(Value: Integer);
    procedure SetTabShowing(Value: Boolean);
    procedure SetTabVisible(Value: Boolean);
    procedure UpdateTabShowing;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
    procedure WriteText(R : TRect; sStyle: TsPaintStyle); override;
  protected
//    procedure CreateParams(var Params: TCreateParams); override;
    procedure DoHide; dynamic;
    procedure DoShow; dynamic;
    procedure ReadState(Reader: TReader); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Invalidate; override;
    property PageControl: TsPageControl read FPageControl write SetPageControl;
    property TabIndex: Integer read GetTabIndex;
  published
//    property Height stored False;
    property Highlighted: Boolean read FHighlighted write SetHighlighted default False;
    property ImageIndex: TImageIndex read FImageIndex write SetImageIndex default 0;
    property Left stored False;
    property PageIndex: Integer read GetPageIndex write SetPageIndex stored False;
    property TabVisible: Boolean read FTabVisible write SetTabVisible default True;
    property Top stored False;
    property Visible stored False;
    property Width stored False;
    property OnHide: TNotifyEvent read FOnHide write FOnHide;
    property OnShow: TNotifyEvent read FOnShow write FOnShow;
  end;

  TsPageControl = class(TCustomTabControl)
  private
    FPages: TList;
    FActivePage: TsTabSheet;
    FNewDockSheet: TsTabSheet;
    FUndockingPage: TsTabSheet;
    FOnDrawTab: TDrawTabEvent;
    FOnGetImageIndex: TTabGetImageEvent;
    FParentShowed : boolean;
    FMainColor: TColor;
    procedure SetCutSize(const Value: integer);
    procedure SetMainColor(const Value: TColor);
    procedure ChangeActivePage(Page: TsTabSheet);
    procedure DeleteTab(Page: TsTabSheet; Index: Integer);
    function GetActivePageIndex: Integer;
    function GetDockClientFromMousePos(MousePos: TPoint): TControl;
    function GetPage(Index: Integer): TsTabSheet;
    function GetPageCount: Integer;
    procedure InsertPage(Page: TsTabSheet);
    procedure InsertTab(Page: TsTabSheet);
    procedure MoveTab(CurIndex, NewIndex: Integer);
    procedure RemovePage(Page: TsTabSheet);
    procedure SetActivePage(Page: TsTabSheet);
    procedure SetActivePageIndex(const Value: Integer);
    procedure UpdateTab(Page: TsTabSheet);
    procedure UpdateTabHighlights;
    procedure CMDesignHitTest(var Message: TCMDesignHitTest); message CM_DESIGNHITTEST;
    procedure CMDialogKey(var Message: TCMDialogKey); message CM_DIALOGKEY;
    procedure CMDockClient(var Message: TCMDockClient); message CM_DOCKCLIENT;
    procedure CMDockNotification(var Message: TCMDockNotification); message CM_DOCKNOTIFICATION;
    procedure CMUnDockClient(var Message: TCMUnDockClient); message CM_UNDOCKCLIENT;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
  protected
    FsStyle : TsStyle;
    FCutSize : integer;
    function CanShowTab(TabIndex: Integer): Boolean; override;
    procedure Change; override;
    procedure DoAddDockClient(Client: TControl; const ARect: TRect); override;
    procedure DockOver(Source: TDragDockObject; X, Y: Integer; State: TDragState; var Accept: Boolean); override;
    procedure DoRemoveDockClient(Client: TControl); override;
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
    function GetImageIndex(TabIndex: Integer): Integer; override;
    function GetPageFromDockClient(Client: TControl): TsTabSheet;
    procedure GetSiteInfo(Client: TControl; var InfluenceRect: TRect; MousePos: TPoint; var CanDock: Boolean); override;
    procedure Loaded; override;
    procedure SetChildOrder(Child: TComponent; Order: Integer); override;
    procedure ShowControl(AControl: TControl); override;
    procedure UpdateActivePage; virtual;

    procedure PaintTabs(C : TCanvas);
    procedure WMPaint (var Message: TWMPaint); message WM_PAINT;
    procedure WMEraseBkGND (var Message: TWMPaint); message WM_ERASEBKGND;
    procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
    procedure WndProc (var Message: TMessage); override;
    procedure DrawTab(C : TCanvas; TabIndex: Integer; const Rect: TRect; Active, DrawParent : boolean);
    procedure DrawBorder(C : TCanvas; Color: TColor; Local : boolean);
    procedure CreateWnd; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Invalidate; override;
    function FindNextPage(CurPage: TsTabSheet; GoForward, CheckTabVisible: Boolean): TsTabSheet;
    procedure SelectNextPage(GoForward: Boolean);
    property CutSize : integer read FCutSize write SetCutSize default 3;
    property MainColor : TColor read FMainColor write SetMainColor;
    property ActivePageIndex: Integer read GetActivePageIndex write SetActivePageIndex;
    property PageCount: Integer read GetPageCount;
    property Pages[Index: Integer]: TsTabSheet read GetPage;
    property Images;
    property MultiLine;
    property RaggedRight;
    property ScrollOpposite;
    property TabPosition;
    property Style;
  published
    property sStyle: TsStyle read FsStyle write FsStyle;
    property ActivePage: TsTabSheet read FActivePage write SetActivePage;
    property Align;
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property OwnerDraw;
    property ParentBiDiMode;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabHeight;
    property TabOrder;
    property TabStop;
    property TabWidth;
    property Visible;
    property OnChange;
    property OnChanging;
    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawTab: TDrawTabEvent read FOnDrawTab write FOnDrawTab;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetImageIndex;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

  TsPageControlEditor = class(TDefaultEditor)
  protected
  public
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

implementation

{ TsTabSheet }

constructor TsTabSheet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Align := alClient;
  ControlStyle := ControlStyle + [csAcceptsControls] + [csOpaque];
  Visible := False;
  FTabVisible := True;
  FHighlighted := False;
end;

destructor TsTabSheet.Destroy;
begin
  if FPageControl <> nil then begin
    if FPageControl.FUndockingPage = Self then FPageControl.FUndockingPage := nil;
    FPageControl.RemovePage(Self);
  end;
  inherited Destroy;
end;

procedure TsTabSheet.DoHide;
begin
  if Assigned(FOnHide) then FOnHide(Self);
end;

procedure TsTabSheet.DoShow;
begin
  if Assigned(FOnShow) then FOnShow(Self);
end;

function TsTabSheet.GetPageIndex: Integer;
begin
  if FPageControl <> nil then
    Result := FPageControl.FPages.IndexOf(Self) else
    Result := -1;
end;

function TsTabSheet.GetTabIndex: Integer;
var
  I: Integer;
begin
  Result := 0;
  if not FTabShowing then begin
    Dec(Result)
  end
  else begin
    for I := 0 to PageIndex - 1 do begin
      if TsTabSheet(FPageControl.FPages[I]).FTabShowing then begin
        Inc(Result);
      end;
    end;
  end;
end;

{
procedure TsTabSheet.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.WindowClass.style := Params.WindowClass.style and not (CS_HREDRAW or CS_VREDRAW);
end;
}
procedure TsTabSheet.ReadState(Reader: TReader);
begin
  inherited ReadState(Reader);
  if Reader.Parent is TsPageControl then PageControl := TsPageControl(Reader.Parent);
end;

procedure TsTabSheet.SetImageIndex(Value: TImageIndex);
begin
  if FImageIndex <> Value then begin
    FImageIndex := Value;
    if FTabShowing then FPageControl.UpdateTab(Self);
  end;
end;

procedure TsTabSheet.SetPageControl(APageControl: TsPageControl);
begin
  if FPageControl <> APageControl then begin
    if FPageControl <> nil then FPageControl.RemovePage(Self);
    Parent := APageControl;
    if APageControl <> nil then APageControl.InsertPage(Self);
  end;
end;

procedure TsTabSheet.SetPageIndex(Value: Integer);
var
  I, MaxPageIndex: Integer;
begin
  if FPageControl <> nil then begin
    MaxPageIndex := FPageControl.FPages.Count - 1;
    if Value > MaxPageIndex then
      raise EListError.CreateResFmt(@SPageIndexError, [Value, MaxPageIndex]);
    I := TabIndex;
    FPageControl.FPages.Move(PageIndex, Value);
    if I >= 0 then FPageControl.MoveTab(I, TabIndex);
  end;
end;

procedure TsTabSheet.SetTabShowing(Value: Boolean);
var
  Index: Integer;
begin
  if FTabShowing <> Value then
    if Value then begin
      FTabShowing := True;
      FPageControl.InsertTab(Self);
    end
    else begin
      Index := TabIndex;
      FTabShowing := False;
      FPageControl.DeleteTab(Self, Index);
    end;
end;

procedure TsTabSheet.SetTabVisible(Value: Boolean);
begin
  if FTabVisible <> Value then
  begin
    FTabVisible := Value;
    UpdateTabShowing;
  end;
end;

procedure TsTabSheet.UpdateTabShowing;
begin
  SetTabShowing((FPageControl <> nil) and FTabVisible);
end;

procedure TsTabSheet.CMTextChanged(var Message: TMessage);
begin
  if FTabShowing then FPageControl.UpdateTab(Self);
end;

procedure TsTabSheet.CMShowingChanged(var Message: TMessage);
begin
  inherited;
  if Showing then begin
    try
      DoShow
    except
      Application.HandleException(Self);
    end;
  end else if not Showing then begin
    try
      DoHide;
    except
      Application.HandleException(Self);
    end;
  end;
end;

procedure TsTabSheet.SetHighlighted(Value: Boolean);
begin
  if not (csReading in ComponentState) then
    SendMessage(PageControl.Handle, TCM_HIGHLIGHTITEM, TabIndex,
      MakeLong(Word(Value), 0));
  FHighlighted := Value;
end;

procedure TsTabSheet.WriteText(R: TRect; sStyle: TsPaintStyle);
begin

end;

procedure TsTabSheet.Invalidate;
begin
  inherited;
  if Assigned(PageControl) and (PageControl.MainColor <> sStyle.Painting.ColorInActive) then begin
    PageControl.MainColor := sStyle.Painting.ColorInActive;
  end;
end;

{ TsPageControl }

constructor TsPageControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  FPages := TList.Create;
  sStyle := TsStyle.Create(Self);
  sStyle.COC := COC_TsPageControl;
  sStyle.Painting.Transparency := 100;
  sStyle.Selection.Mode := [];
  FCutSize := 3;
end;

procedure TsPageControl.Invalidate;
begin
  if (TabHeight = 0) and Assigned(Parent) then TabHeight := 18;
  if (TabWidth = 0) and Assigned(Parent) then TabWidth := 96;
  inherited;
end;

procedure TsPageControl.WMPaint(var Message: TWMPaint);
var
  SaveIndex: Integer;
  DC: HDC;
  PS: TPaintStruct;
  aRect : TRect;
begin
//  if sStyle.Flat and not (csDestroying in ComponentState) and not (csLoading in ComponentState) then begin
    DC := Message.DC;
    if DC = 0 then DC := BeginPaint(Handle, PS);
    SaveIndex := SaveDC(DC);

    sStyle.FCacheBmp.Height := Height;
    sStyle.FCacheBmp.Width := Width;

    if PageCount > 0 then begin
      DrawBorder(sStyle.FCacheBmp.Canvas, MainColor, False);
      PaintTabs(sStyle.FCacheBmp.Canvas);
    end
    else begin
      if sStyle.GetParentCache <> nil then begin
        FadeRect(sStyle.GetParentCache.Canvas,
                   Rect(Left, Top, Left + Width, Top + Height),
                   sStyle.FCacheBmp.Canvas, Point(0, 0), sStyle.Painting.Transparency, sStyle.Painting.ColorInActive, 0, ssRectangle);
      end
      else begin
        sStyle.FCacheBmp.Canvas.Pen.Style := psClear;
        sStyle.FCacheBmp.Canvas.Brush.Style := bsSolid;
        sStyle.FCacheBmp.Canvas.Brush.Color := sStyle.Painting.ColorInActive;
        sStyle.FCacheBmp.Canvas.Rectangle(aRect);
      end;
      sStyle.PaintBorder(sStyle.FCacheBmp.Canvas.Handle, ClientRect, False);
    end;

    if Message.DC = 0
      then aRect := PS.rcPaint
      else aRect := ClientRect;
    sStyle.CopyFromCache(DC, aRect.Left, aRect.Top, aRect.Right, aRect.Bottom);
    RestoreDC(DC, SaveIndex);
//    PaintControls(DC, Self);
    if Message.DC = 0 then EndPaint(Handle, PS);
    if (csDesigning in ComponentState) and (PageCount > 0) and ActivePage.Visible then ActivePage.Repaint;
    Message.Result := 1;
//  end;
end;

procedure TsPageControl.DrawBorder(C : TCanvas; Color: TColor; Local : boolean);
var
  R : TRect;
  W : integer;
begin
  W := 5;
  R := Rect(0, TabHeight + 4, Width + 1, Height + 1);

  InflateRect(R, -1, -1);
  if Local then
    DrawRectangleOnDC(Canvas.Handle, R, Color, Color, W)
  else begin
    C.Brush.Style := bsSolid;
    C.Pen.Style := psClear;
    C.Brush.Color := Color;
    C.FillRect(R);
    sStyle.PaintBorder(C.Handle, Rect(0, TabHeight + 4, Width, Height), False);
  end;
end;

procedure TsPageControl.Change;
var
  Form: TCustomForm;
begin
  if not (csDestroying in ComponentState) and
     not (csLoading in ComponentState) then begin
    UpdateActivePage;
    if csDesigning in ComponentState then begin
      Form := GetParentForm(Self);
      if (Form <> nil) and (Form.Designer <> nil) then Form.Designer.Modified;
    end;
    inherited Change;
  end;
end;

procedure TsPageControl.WMEraseBkGND(var Message: TWMPaint);
begin
  Message.Result := 1;
end;

procedure TsPageControl.WMNCPaint(var Message: TMessage);
begin
  inherited;
  if not FParentShowed and False{not (csDestroying in ComponentState)} then begin
{
    FParentShowed := True;
    Parent.Perform(WM_PAINT, Canvas.Handle, 0);
    if sStyle.GetParentCache <> nil then
      FadeRect(sStyle.GetParentCache.Canvas, Rect(Left, Top, Left + Width, Top + Height), sStyle.FCacheBmp.Canvas, Point(0, 0), 100, clSilver, 0, ssRectangle);
    if PageCount > 0 then PaintTabs(sStyle.FCacheBmp.Canvas);
    sStyle.CopyFromCache(Canvas.Handle, 0, 0, Width, Height);
}    
  end;
{
  if PageCount = 0 then begin
      FadeRect(sStyle.GetParentCache.Canvas,
                   Rect(Left, Top, Left + Width, Top + Height),
                   sStyle.FCacheBmp.Canvas, Point(0, 0), 0, clSilver, 0, ssRectangle);
      sStyle.PaintBorder(sStyle.FCacheBmp.Canvas.Handle, 0, 0, Width, Height, False);
  end;
}
end;

procedure TsPageControl.SetCutSize(const Value: integer);
begin
  if FCutSize <> Value then begin
    if Value < 0 then FCutSize := 0
    else if Value > TabHeight then FCutSize := TabHeight
    else FCutSize := Value;
    Invalidate;
  end;
end;

procedure TsPageControl.WndProc(var Message: TMessage);
begin
//  if not (csDestroying in ComponentState) and
//     not (csLoading in ComponentState) then
  sStyle.WndProc(Message);
  if Message.Result <> 1 then
    inherited;
end;

function TsPageControl.GetPage(Index: Integer): TsTabSheet;
begin
  Result := FPages[Index];
end;

function TsPageControl.GetPageCount: Integer;
begin
  Result := FPages.Count;
end;

function TsPageControl.GetActivePageIndex: Integer;
begin
  if ActivePage <> nil then
    Result := ActivePage.GetPageIndex
  else
    Result := -1;
end;

procedure TsPageControl.SetActivePage(Page: TsTabSheet);
begin
  if (Page <> nil) and (Page.PageControl <> Self) then Exit;
  ChangeActivePage(Page);
  if Page = nil then
    TabIndex := -1
  else if Page = FActivePage then
    TabIndex := Page.TabIndex;
end;

procedure TsPageControl.SetActivePageIndex(const Value: Integer);
begin
  if (Value > -1) and (Value < PageCount) then
    ActivePage := Pages[Value]
  else
    ActivePage := nil;
end;

procedure TsPageControl.RemovePage(Page: TsTabSheet);
var
  NextSheet: TsTabSheet;
begin
  NextSheet := FindNextPage(Page, True, not (csDesigning in ComponentState));
  if NextSheet = Page then NextSheet := nil;
  Page.SetTabShowing(False);
  Page.FPageControl := nil;
  FPages.Remove(Page);
  SetActivePage(NextSheet);
end; 

procedure TsPageControl.UpdateTab(Page: TsTabSheet);
begin
  Tabs[Page.TabIndex] := Page.Caption;
end;

procedure TsPageControl.DeleteTab(Page: TsTabSheet; Index: Integer);
var
  UpdateIndex: Boolean;
begin
  UpdateIndex := Page = ActivePage;
  Tabs.Delete(Index);
  if UpdateIndex then
  begin
    if Index >= Tabs.Count then
      Index := Tabs.Count - 1;
    TabIndex := Index;
  end;
  UpdateActivePage;
end;

procedure TsPageControl.InsertPage(Page: TsTabSheet);
begin
  FPages.Add(Page);
  Page.FPageControl := Self;
  Page.ControlStyle := [csAcceptsControls, csClickEvents,
    {csSetCaption,} csDoubleClicks{, csReplicatable}];
  Page.UpdateTabShowing;
end;

procedure TsPageControl.InsertTab(Page: TsTabSheet);
begin
  Tabs.InsertObject(Page.TabIndex, Page.Caption, Page);
  UpdateActivePage;
end;

procedure TsPageControl.MoveTab(CurIndex, NewIndex: Integer);
begin
  Tabs.Move(CurIndex, NewIndex);
end;

procedure TsPageControl.ChangeActivePage(Page: TsTabSheet);
var
  ParentForm: TCustomForm;
begin
  if FActivePage <> Page then begin
{
    if Assigned(ActivePage) and (ActivePage.GetPageIndex >= 0) then begin
      DrawTab(Canvas, FActivePage.GetPageIndex, TabRect(FActivePage.GetPageIndex), False, True);
    end;
    DrawTab(Canvas, Page.GetPageIndex, TabRect(Page.GetPageIndex), True, False);
}
    ParentForm := GetParentForm(Self);
    if (ParentForm <> nil) and (FActivePage <> nil) and
          FActivePage.ContainsControl(ParentForm.ActiveControl) then begin
      ParentForm.ActiveControl := FActivePage;
      if ParentForm.ActiveControl <> FActivePage then begin
        TabIndex := FActivePage.TabIndex;
        Exit;
      end;
    end;
    if Page <> nil then begin
      Page.BringToFront;
      Page.Visible := True;
      if (ParentForm <> nil) and (FActivePage <> nil) and (ParentForm.ActiveControl = FActivePage) then
        if Page.CanFocus then
          ParentForm.ActiveControl := Page else
          ParentForm.ActiveControl := Self;
    end;
    if FActivePage <> nil then FActivePage.Visible := False;
    FActivePage := Page;

    PaintTabs(sStyle.FCacheBmp.Canvas);
    sStyle.CopyFromCache(Canvas.Handle, 0, 0, Width, Height);//TabHeight + 6);

    if (ParentForm <> nil) and (FActivePage <> nil) and
      (ParentForm.ActiveControl = FActivePage) then
      FActivePage.SelectFirst;
  end;
end;

function TsPageControl.FindNextPage(CurPage: TsTabSheet; GoForward,
  CheckTabVisible: Boolean): TsTabSheet;
var
  I, StartIndex: Integer;
begin
  if FPages.Count <> 0 then
  begin
    StartIndex := FPages.IndexOf(CurPage);
    if StartIndex = -1 then
      if GoForward then StartIndex := FPages.Count - 1 else StartIndex := 0;
    I := StartIndex;
    repeat
      if GoForward then
      begin
        Inc(I);
        if I = FPages.Count then I := 0;
      end else
      begin
        if I = 0 then I := FPages.Count;
        Dec(I);
      end;
      Result := FPages[I];
      if not CheckTabVisible or Result.TabVisible then Exit;
    until I = StartIndex;
  end;
  Result := nil;
end;

procedure TsPageControl.UpdateActivePage;
begin
  if TabIndex >= 0 then
    SetActivePage(TsTabSheet(Tabs.Objects[TabIndex]))
  else
    SetActivePage(nil);
end;

destructor TsPageControl.Destroy;
var
  I: Integer;
begin
  for I := 0 to FPages.Count - 1 do TsTabSheet(FPages[I]).FPageControl := nil;
  FreeAndNil(FPages);
  FreeAndNil(FsStyle);
  inherited Destroy;
end;

procedure TsPageControl.CMDesignHitTest(var Message: TCMDesignHitTest);
var
  HitIndex: Integer;
  HitTestInfo: TTCHitTestInfo;
begin
  HitTestInfo.pt := SmallPointToPoint(Message.Pos);
  HitIndex := SendMessage(Handle, TCM_HITTEST, 0, Longint(@HitTestInfo));
  if (HitIndex >= 0) and (HitIndex <> TabIndex) then Message.Result := 1;
end;

procedure TsPageControl.CMDialogKey(var Message: TCMDialogKey);
begin
  if (Focused or Windows.IsChild(Handle, Windows.GetFocus)) and
    (Message.CharCode = VK_TAB) and (GetKeyState(VK_CONTROL) < 0) then
  begin
    SelectNextPage(GetKeyState(VK_SHIFT) >= 0);
    Message.Result := 1;
  end else
    inherited;
end;

procedure TsPageControl.CMDockClient(var Message: TCMDockClient);
var
  IsVisible: Boolean;
  DockCtl: TControl;
begin
  Message.Result := 0;
  FNewDockSheet := TsTabSheet.Create(Self);
  try
    try
      DockCtl := Message.DockSource.Control;
      if DockCtl is TCustomForm then
        FNewDockSheet.Caption := TCustomForm(DockCtl).Caption;
      FNewDockSheet.PageControl := Self;
      DockCtl.Dock(Self, Message.DockSource.DockRect);
    except
      FNewDockSheet.Free;
      raise;
    end;
    IsVisible := DockCtl.Visible;
    FNewDockSheet.TabVisible := IsVisible;
    if IsVisible then ActivePage := FNewDockSheet;
    DockCtl.Align := alClient;
  finally
    FNewDockSheet := nil;
  end;
end;

procedure TsPageControl.CMDockNotification(var Message: TCMDockNotification);
var
  I: Integer;
  S: string;
  Page: TsTabSheet;
begin
  Page := GetPageFromDockClient(Message.Client);
  if Page <> nil then
    case Message.NotifyRec.ClientMsg of
      WM_SETTEXT:
        begin
          S := PChar(Message.NotifyRec.MsgLParam);
          { Search for first CR/LF and end string there }
          for I := 1 to Length(S) do
            if S[I] in [#13, #10] then
            begin
              SetLength(S, I - 1);
              Break;
            end;
          Page.Caption := S;
        end;
      CM_VISIBLECHANGED:
        Page.TabVisible := Boolean(Message.NotifyRec.MsgWParam);
    end;
  inherited;
end;

procedure TsPageControl.CMUnDockClient(var Message: TCMUnDockClient);
var
  Page: TsTabSheet;
begin
  Message.Result := 0;
  Page := GetPageFromDockClient(Message.Client);
  if Page <> nil then
  begin
    FUndockingPage := Page;
    Message.Client.Align := alNone;
  end;
end;

procedure TsPageControl.WMLButtonDblClk(var Message: TWMLButtonDblClk);
var
  DockCtl: TControl;
begin
  inherited;
  DockCtl := GetDockClientFromMousePos(SmallPointToPoint(Message.Pos));
  if DockCtl <> nil then DockCtl.ManualDock(nil, nil, alNone);
end;

procedure TsPageControl.WMLButtonDown(var Message: TWMLButtonDown);
var
  DockCtl: TControl;
begin
  inherited;
  DockCtl := GetDockClientFromMousePos(SmallPointToPoint(Message.Pos));
  if (DockCtl <> nil) and (Style = tsTabs) then DockCtl.BeginDrag(False);
end;

procedure TsPageControl.SelectNextPage(GoForward: Boolean);
var
  Page: TsTabSheet;
begin
  Page := FindNextPage(ActivePage, GoForward, True);
  if (Page <> nil) and (Page <> ActivePage) and CanChange then
  begin
    TabIndex := Page.TabIndex;
    Change;
  end;
end;

function TsPageControl.GetPageFromDockClient(Client: TControl): TsTabSheet;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to PageCount - 1 do
  begin
    if (Client.Parent = Pages[I]) and (Client.HostDockSite = Self) then
    begin
      Result := Pages[I];
      Exit;
    end;
  end;
end;

function TsPageControl.GetDockClientFromMousePos(
  MousePos: TPoint): TControl;
var
  i, HitIndex: Integer;
  HitTestInfo: TTCHitTestInfo;
  Page: TsTabSheet;
begin
  Result := nil;
  if DockSite then
  begin
    HitTestInfo.pt := MousePos;
    HitIndex := SendMessage(Handle, TCM_HITTEST, 0, Longint(@HitTestInfo));
    if HitIndex >= 0 then
    begin
      Page := nil;
      for i := 0 to HitIndex do
        Page := FindNextPage(Page, True, True);
      if (Page <> nil) and (Page.ControlCount > 0) then
      begin
        Result := Page.Controls[0];
        if Result.HostDockSite <> Self then Result := nil;
      end;
    end;
  end;
end;

procedure TsPageControl.DrawTab(C : TCanvas; TabIndex: Integer; const Rect: TRect; Active, DrawParent : boolean);
var
  rText, rParent, aRect : TRect;
begin
  rParent := Rect;
  inc(rParent.Right, 4);
  inc(rParent.Bottom, 1);
  dec(rParent.Top, 2);
  dec(rParent.Left, 2);

  if Active and sStyle.Shadow.Enabled then begin
    aRect := Rect;

    FadeRect(C,
             Classes.Rect(aRect.Right - sStyle.Shadow.Offset,
                          aRect.Top + sStyle.Shadow.Offset,
                          aRect.Right + sStyle.Shadow.Offset + 6,
                          aRect.Bottom + sStyle.Shadow.Offset + 2),
             C, Point(aRect.Right - sStyle.Shadow.Offset, aRect.Top + sStyle.Shadow.Offset),
             sStyle.Shadow.Transparency, sStyle.Shadow.Color, sStyle.Shadow.Blur, ssRectangle);

  end;

  // Painting of parent rect for BG
  OffsetRect(rParent, Left, Top);
  if sStyle.GetParentCache <> nil then
    FadeRect(sStyle.GetParentCache.Canvas, rParent, C, Point(Rect.Left - 2, Rect.Top - 2), 100, clSilver, 0, ssRectangle)
  else begin
    C.Brush.Style := bsSolid;
    C.Pen.Style := psClear;
    C.Brush.Color := Pages[TabIndex].sStyle.Painting.ColorInActive;
    C.FillRect(Classes.Rect(Rect.Left - 2, Rect.Top - 2, Rect.Right + 2, Rect.Bottom + 2));
  end;

  C.Brush.Style := bsSolid;
  C.Pen.Style := psSolid;
  C.Brush.Color := Pages[TabIndex].sStyle.Painting.ColorInActive;
  C.Pen.Color := Pages[TabIndex].sStyle.Painting.ColorInActive;
  aRect := Rect;
  rText := Rect;

  if Active then begin
    dec(aRect.Top, 2);
    dec(aRect.Left, 2);
    inc(aRect.Right, 4);
    C.RoundRect(aRect.Left, aRect.Top,
                     aRect.Right, aRect.Bottom + 2,
                     4, 4);

    DrawBorder(C, Pages[TabIndex].sStyle.Painting.ColorInActive, False);

    C.Brush.Style := bsSolid;
    C.Brush.Color := Pages[TabIndex].sStyle.Painting.ColorInActive;
    C.Pen.Style := psClear;
    C.Rectangle(aRect.Left, aRect.Bottom + 1,
                     aRect.Right + 2, aRect.Bottom + 4);

    // Top border
    C.Pen.Style := psSolid;
    C.Brush.Style := bsClear;
    C.Pen.Color := sStyle.Painting.ColorBorderTop;
    C.PolyLine([Point(aRect.Left, aRect.Bottom + 2),
                     Point(aRect.Left, aRect.Top + 2),
                     Point(aRect.Left + 2, aRect.Top),
                     Point(aRect.Right - 2, aRect.Top)]);


    // Bottom border
    C.Pen.Color := sStyle.Painting.ColorBorderBottom;
    C.PolyLine([Point(aRect.Right - 2, aRect.Top),
                     Point(aRect.Right, aRect.Top + 2),
                     Point(aRect.Right, aRect.Bottom + 3)]);

    if Focused then
      C.DrawFocusRect(rText);

  end
  else begin
    inc(aRect.Bottom, 4);
    dec(aRect.Left, 2);
    inc(aRect.Right, 2);

    inc(rText.Top, 2);

    if sStyle.GetParentCache <> nil then

      FadeRect(sStyle.GetParentCache.Canvas,
                     aRect, C,
                     Point(aRect.Left, aRect.Top),
                     Pages[TabIndex].sStyle.Painting.Transparency, // + SumTrans(tr, Shadow.Transparency, 100),
                     Pages[TabIndex].sStyle.Painting.ColorInActive, 1, ssRectangle);

    C.Pen.Color := ColorToRGB(clBtnShadow);//sStyle.Painting.ColorBorderBottom;
    C.Brush.Style := bsClear;
    C.RoundRect(aRect.Left, aRect.Top, aRect.Right, aRect.Bottom, 2, 2);
  end;

  if not OwnerDraw then begin
    WriteText(C, PChar(Pages[TabIndex].Caption), True,
                           rText, DT_CENTER or ta_CENTER);
  end
  else begin
    if Assigned(FOnDrawTab) then FOnDrawTab(Self, TabIndex, aRect, Active);
  end;
end;

procedure TsPageControl.PaintTabs(C : TCanvas);
var
  i : integer;
begin
  if sStyle.GetParentCache <> nil then
    FadeRect(sStyle.GetParentCache.Canvas,
                   Rect(Left, Top, Left + Width, Top + TabHeight + 4),
                   C, Point(0, 0), 100, clSilver, 0, ssRectangle)
  else begin
    C.Brush.Style := bsSolid;
    C.Pen.Style := psClear;
    C.Brush.Color := sStyle.Painting.ColorInActive;
    C.FillRect(Rect(0, 0, Width, TabHeight + 4));
  end;
  for i := 0 to PageCount - 1 do begin
    if Pages[i] <> ActivePage then
      DrawTab(C, i, TabRect(i), False, False);
  end;
  DrawTab(C, ActivePage.GetPageIndex, TabRect(ActivePage.GetPageIndex), True, False);
end;


procedure TsPageControl.SetMainColor(const Value: TColor);
begin
  FMainColor := Value;
  Invalidate;
end;

function TsPageControl.CanShowTab(TabIndex: Integer): Boolean;
begin
  Result := TsTabSheet(FPages[TabIndex]).Enabled;
end;

procedure TsPageControl.DoAddDockClient(Client: TControl; const ARect: TRect);
begin
  if FNewDockSheet <> nil then Client.Parent := FNewDockSheet;
end;

procedure TsPageControl.DockOver(Source: TDragDockObject; X, Y: Integer;
  State: TDragState; var Accept: Boolean);
var
  R: TRect;
begin
  GetWindowRect(Handle, R);
  Source.DockRect := R;
  DoDockOver(Source, X, Y, State, Accept);
end;

procedure TsPageControl.DoRemoveDockClient(Client: TControl);
begin
  if (FUndockingPage <> nil) and not (csDestroying in ComponentState) then
  begin
    SelectNextPage(True);
    FUndockingPage.Free;
    FUndockingPage := nil;
  end;
end;

procedure TsPageControl.GetChildren(Proc: TGetChildProc; Root: TComponent);
var
  I: Integer;
begin
  for I := 0 to FPages.Count - 1 do Proc(TComponent(FPages[I]));
end;

function TsPageControl.GetImageIndex(TabIndex: Integer): Integer;
var
  I,
  Visible,
  NotVisible: Integer;
begin
  if Assigned(FOnGetImageIndex) then
    Result := inherited GetImageIndex(TabIndex) else
    begin
     { For a PageControl, TabIndex refers to visible tabs only. The control
     doesn't store }
      Visible := 0;
      NotVisible := 0;
      for I := 0 to FPages.Count - 1 do
      begin
        if not GetPage(I).TabVisible then Inc(NotVisible)
        else Inc(Visible);
        if Visible = TabIndex + 1 then Break;
      end;
      Result := GetPage(TabIndex + NotVisible).ImageIndex;
    end;
end;

procedure TsPageControl.GetSiteInfo(Client: TControl;
  var InfluenceRect: TRect; MousePos: TPoint; var CanDock: Boolean);
begin
  CanDock := GetPageFromDockClient(Client) = nil;
  inherited GetSiteInfo(Client, InfluenceRect, MousePos, CanDock);
end;

procedure TsPageControl.Loaded;
begin
  inherited Loaded;
  UpdateTabHighlights;
end;

procedure TsPageControl.SetChildOrder(Child: TComponent; Order: Integer);
begin
  TTabSheet(Child).PageIndex := Order;
end;

procedure TsPageControl.ShowControl(AControl: TControl);
begin
  if (AControl is TsTabSheet) and (TsTabSheet(AControl).PageControl = Self) then
    SetActivePage(TsTabSheet(AControl));
  inherited ShowControl(AControl);
end;

procedure TsPageControl.UpdateTabHighlights;
var
  I: Integer;
begin
  for I := 0 to PageCount - 1 do
    Pages[I].SetHighlighted(Pages[I].FHighlighted);
end;

procedure TsPageControl.CreateWnd;
begin
  inherited;
  sStyle.CreateWnd;
end;

{ TsPageControlEditor }

procedure TsPageControlEditor.ExecuteVerb(Index: Integer);
var
  NewPage: TsTabSheet;
begin
  case Index of
    0:  begin  //New Page
          NewPage := TsTabSheet.Create(Designer.GetRoot);
          NewPage.Parent := (Component as TsPageControl);
          NewPage.PageControl := (Component as TsPageControl);
          NewPage.Caption := Designer.UniqueName('sTabSheet');
          NewPage.Name := NewPage.Caption;
        end;
    1:  begin  //Delete Page
          with (Component as TsPageControl) do begin
            NewPage := ActivePage;
//            NewPage.PageControl := nil;
            NewPage.Free;
          end;
        end;
    2:  begin  //Next Page
          (Component as TsPageControl).FindNextPage((Component as TsPageControl).ActivePage,True,False);
        end;
    3:  begin  //  Previous Page
          (Component as TsPageControl).FindNextPage((Component as TsPageControl).ActivePage,False,False);
        end;
  end;
  if Designer <> nil then Designer.Modified;
end;

function TsPageControlEditor.GetVerb(Index: Integer): string;
begin
  case Index of
    0:  result := 'New Page';
    1:  result := 'Delete Page';
    2:  result := 'Next Page';
    3:  result := 'Previous Page';
  end;
end;

function TsPageControlEditor.GetVerbCount: Integer;
begin
  result := 4;
end;

end.
